import unittest


class SphericalProjectionTestCase(unittest.TestCase):

    @unittest.expectedFailure
    def test_equal_area(self):
        raise NotImplementedError()

    @unittest.expectedFailure
    def test_equal_angle(self):
        raise NotImplementedError()

    @unittest.expectedFailure
    def test__project(self):
        raise NotImplementedError()



"""
model_sampling_strike_slip
**************************

Example extension module for generating strike slip events
"""
import numpy as np
from MTfit.convert import Tape_MT6


def random_strike_slip(self, n_samples):
    kappa = np.random.rand(n_samples)*2*np.pi
    h = np.ones(n_samples,)
    sigma = np.random.rand(n_samples)*np.pi-np.pi/2
    return Tape_MT6(np.zeros(n_samples), np.zeros(n_samples), kappa, h, sigma)



"""
argparser_defaults.py
***********************

Dictionaries of MTfit and MTPlot argparser defaults and
default types

"""


# **Restricted:  For Non-Commercial Use Only**
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.

from distutils.version import StrictVersion

from matplotlib import __version__ as matplotlib_version
#
# MTfit defaults
#
# Parameters for command line argument parser
# Defaults are set in this dictionary and updated by site and user defaults

MTFIT_PARSER_DEFAULTS = {
    'data_file': False,
    'location_pdf_file_path': False,
    'algorithm': 'time',
    'single_threaded': False,
    'number_workers': 0,
    'number_stations': 0,
    'memory': 8,
    'double-couple': False,
    'compare_constrained': False,
    'nstations': 0,
    'number_location_samples': 0,
    'disk_sample': False,
    'no_file_safe': False,
    'inversion_options': False,
    'output_file': False,
    'max_iteration_samples': 6000000,
    'max_time': 600,
    'multiple_events': False,
    'relative_amplitude': False,
    'marginalise_relative': False,
    'recover': False,
    'mpi': False,
    'benchmark': False,
    'test': False,
    'quality': False,
    'quality_check_value': 10,
    'debug': False,
    'diagnostics': False,
    'no_dist': False,
    'learning_length': 5000,
    'warnings': 'default',
    'acceptance_rate_window': 500,
    'initial_sampling': 'grid',
    'jump_probability': 0.01,
    'mcmc_min_number_initialisation_samples': 5000,
    'transdmcmc_min_number_initialisation_samples': 5000,
    'min_number_check_samples': 30000,
    'mcmc_max_acceptance_rate': 0.5,
    'mcmc_min_acceptance_rate': 0.3,
    'transdmcmc_max_acceptance_rate': 0.2,
    'transdmcmc_min_acceptance_rate': 0.05,
    'mcmc_chain_length': 10000,
    'transdmcmc_chain_length': 100000,
    'data_extension': 'inv',
    'location_extension': 'scatangle',
    'min_number_intersections': 2,
    'verbosity': 0,
    'nodes': 1,
    'ppn': 8,
    'pmem': 2,
    'walltime': '24:00:00',
    'queue': 'batch',
    'email': False,
    'email_options': 'bae',
    'output_format': 'matlab',
    'results_format': 'full_pdf',
    'no_normalise': False,
    'convert': False,
    'discard': 10000,
    'dc_prior': 0.5,
    'sampling': False,
    'sample_distribution': False,
    'sampling_prior': False,
    'mpi_output': False,
    'c_generate': False,
    'combine_mpi_output': False,
    'relative_loop': False
}

# Default types for the ARGPARSER_DEFAULTS structure, used for checking
# option types
MTFIT_PARSER_DEFAULT_TYPES = {
    'data_file': [str, bool],
    'location_pdf_file_path': [str, bool],
    'algorithm': [str],
    'single_threaded': [bool],
    'number_workers': [int],
    'number_stations': [int],
    'memory': [float],
    'double-couple': [bool],
    'compare_constrained': [bool],
    'nstations': [int],
    'number_location_samples': [int],
    'disk_sample': [bool],
    'no_file_safe': [bool],
    'inversion_options': [str, bool],
    'output_file': [str, bool],
    'max_iteration_samples': [int],
    'max_time': [float, int],
    'multiple_events': [bool],
    'relative_amplitude': [bool],
    'marginalise_relative': [bool],
    'recover': [bool],
    'mpi': [bool],
    'benchmark': [bool],
    'test': [bool],
    'quality': [bool],
    'quality_check_value': [float],
    'debug': [bool],
    'diagnostics': [bool],
    'mcmc_min_acceptance_rate': [float],
    'mcmc_max_acceptance_rate': [float],
    'no_dist': [bool],
    'learning_length': [int],
    'warnings': 'default',
    'acceptance_rate_window': [int],
    'initial_sampling': [str],
    'jump_probability': [float],
    'min_number_check_samples': [int],
    'mcmc_min_number_initialisation_samples': [int],
    'transdmcmc_min_number_initialisation_samples': [int],
    'data_extension': [str],
    'location_extension': [str],
    'min_number_intersections': [int],
    'verbosity': [int],
    'nodes': [int],
    'ppn': [int],
    'pmem': [float],
    'walltime': [str],
    'queue': [str],
    'email': [str, bool],
    'email_options': [str, bool],
    'transdmcmc_min_acceptance_rate': [float],
    'transdmcmc_max_acceptance_rate': [float],
    'mcmc_chain_length': [int],
    'transdmcmc_chain_length': [int],
    'output_format': [str],
    'results_format': [str],
    'no_normalise': [bool],
    'discard': [float],
    'convert': [bool],
    'dc_prior': [float],
    'sampling': [bool, str],
    'sampling_prior': [bool, str],
    'sample_distribution': [bool, str],
    'mpi_output': [bool],
    'c_generate': [bool],
    'combine_mpi_output': [bool],
    'relative_loop': [bool]
}

# MTplot defaults are set in this dictionary and updated by site and user
# defaults
DEFAULT_HIST_COLORMAP = 'viridis'
if StrictVersion(matplotlib_version) < StrictVersion('1.5.0'):
    DEFAULT_HIST_COLORMAP = 'CMRmap'
DEFAULT_AMP_COLORMAP = 'bwr'

MTPLOT_PARSER_DEFAULTS = {
    'plot_type': 'beachball',
    # Uses DEFAULT_AMP_COLORMAP or DEFAULT_HIST_COLORMAP as set above
    'colormap': False,
    'fontsize': 11,
    'linewidth': 1,
    'text': False,
    'resolution': 500,
    'bins': 100,
    'fault_plane': False,
    'nodal_line': True,
    'TNP': False,
    'markersize': 10,
    'station_markersize': 10,
    'show_max_likelihood': False,
    'show_mean': False,
    'grid_lines': False,
    'color': 'purple',
    'type_label': False,
    'hex_bin': False,
    'projection': 'equalarea',
    'save_file': '',
    'save_dpi': 200,
    'hide': False
}

# Default types for the MTplot parser structure, used for checking option types
MTPLOT_PARSER_DEFAULT_TYPES = {
    'plot_type': [str],
    'colormap': [str],
    'fontsize': [int],
    'linewidth': [int],
    'text': [bool],
    'resolution': [int],
    'bins': [int],
    'fault_plane': [bool],
    'nodal_line': [bool],
    'TNP': [bool],
    'markersize': [int],
    'station_markersize': [int],
    'show_max_likelihood': [bool],
    'show_mean': [bool],
    'grid_lines': [bool],
    'color': [str],
    'type_label': [bool],
    'hex_bin': [bool],
    'projection': [str],
    'save_file': [str],
    'save_dpi': [int],
    'hide': [bool],
}



"""
sampling.py
***********
module containing basic Sample object for storing samples from the source pdf.
"""


# **Restricted:  For Non-Commercial Use Only**
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.

import os
import sys
import shutil


import numpy as np


from .probability import LnPDF
from .probability import dkl_estimate
from .probability.probability import _6sphere_prior
from .utilities.file_io import convert_keys_to_unicode
from .utilities.file_io import unique_columns
from .convert import output_convert

# Because long doesn't exist in python 3 our isinstance tests fail
if sys.version_info.major >= 3:
    long = int


__all__ = ['Sample', 'FileSample', 'ln_bayesian_evidence', '_convert', '_6sphere_prior']


class Sample(object):
    """Sample object for storing source pdf samples."""

    def __init__(self, initial_sample_size=100000, number_events=1, prior=_6sphere_prior):
        """
        Sample initialisation

        Args
            initial_sample_size:[100000] Initial size for MT vector
            number_events:[1] Number of events in sample (allows for sampling multiple event joint PDF)
            prior:[6sphere]

        Returns
            Sample object

        """
        self.moment_tensors = np.matrix(np.zeros((number_events*6, initial_sample_size)))
        self.n = 0
        self.ln_pdf = LnPDF()
        self.number_events = number_events
        self._initial_sample_size = initial_sample_size
        self._i = 0
        self._prior = prior

    def append(self, moment_tensors, ln_pdf, n, scale_factor=False, extensions_scale_factor=False):
        """
        Appends new samples to the Sample object, extending the MT array if the limit is reached.

        Args
            moment_tensors: numpy matrix of moment tensors to append.
            ln_pdf: numpy matrix/array or PDF object containing ln_pdf samples corresponding to the moment tensors in moment_tensors (must be same length).
            n: number of tried samples (including zero probability samples).
            scale_factor:[False] scale_factor estimates for relative amplitude inversion.
            extensions_scale_factor:[False] scale_factor estimates for extensions carrying out relative inversion.

        Returns
            None

        """
        # If scale_factor is passed as an argument, and no scale_factor data stored, create the array
        if not isinstance(scale_factor, bool) and not hasattr(self, 'scale_factor'):
            self.scale_factor = np.array([])
        if isinstance(extensions_scale_factor, dict) and len(extensions_scale_factor) and not hasattr(self, 'extensions_scale_factor'):
            self.extensions_scale_factor = {}
        # Check if moment tensors are a list (multiple events) and convert to the correct format
        if isinstance(moment_tensors, list) and isinstance(moment_tensors[0], np.ndarray):
            moment_tensors = np.array(moment_tensors)
            moment_tensors = np.matrix(moment_tensors.reshape(moment_tensors.shape[0]*moment_tensors.shape[1], moment_tensors.shape[2]))
        # Check moment tensor and PDF shape
        if isinstance(ln_pdf, (float, long, int, np.float64, np.float32)) and ((isinstance(moment_tensors, np.ndarray) and moment_tensors.shape[1] != 1) or (isinstance(moment_tensors, list) and max([mt.shape[1] for mt in moment_tensors]) != 1)):
            raise ValueError('Moment Tensor shape must be 6x1 when using a float ln_pdf')
        elif not isinstance(ln_pdf, (float, long, int, np.float64, np.float32)) and moment_tensors.shape[1] != ln_pdf.shape[1]:
            raise ValueError('Moment Tensor shape[2] and ln_pdf shape[2] must be the same')
        # Add number of samples (n) to total number of samples (self.n)
        self.n += n
        # Convert ln_pdf to ln_pdf form
        if isinstance(ln_pdf, (np.ndarray, float, int, np.float64, np.float32)):
            ln_pdf = LnPDF(ln_pdf)
        # Get non-zero probability samples
        moment_tensors = moment_tensors[:, ln_pdf.nonzero()]
        # Check scale factor
        if not isinstance(scale_factor, bool):
            if not isinstance(scale_factor, dict):  # single event
                scale_factor = scale_factor[ln_pdf.nonzero()]
        if isinstance(extensions_scale_factor, dict) and len(extensions_scale_factor):
            for key in extensions_scale_factor.keys():
                extensions_scale_factor[key] = extensions_scale_factor[key][ln_pdf.nonzero()]
        # Check if there are non-zero moment tensor samples
        if moment_tensors.shape[0]*moment_tensors.shape[1] > 0:  # Has samples
            if len(ln_pdf.nonzero()) != moment_tensors.shape[1]:
                raise ValueError('Moment Tensor shape[2] and ln_pdf shape[2] must be the same')
            # Check if moment tensor sampling array has enough spare elements
            size_check = self.moment_tensors.shape[1]-self._i > moment_tensors.shape[1]
            # If size-check fails, append new zero samples to the array
            while not size_check:
                self.moment_tensors = np.append(self.moment_tensors, np.zeros((self.number_events*6, self._initial_sample_size)),
                                                axis=1)
                size_check = self.moment_tensors.shape[1]-self._i > moment_tensors.shape[1]
            # Set moment tensor samples in array
            self.moment_tensors[:, self._i:self._i+moment_tensors.shape[1]] = moment_tensors
            # Handle scale factors
            if not isinstance(scale_factor, bool):
                # Scale factor dict containing mu and s
                self.scale_factor = np.append(self.scale_factor, scale_factor)
            if isinstance(extensions_scale_factor, dict) and len(extensions_scale_factor):
                # Scale factor dict containing mu and s
                for key in extensions_scale_factor.keys():
                    if key in self.extensions_scale_factor.keys():
                        self.extensions_scale_factor[key] = np.append(self.extensions_scale_factor[key], extensions_scale_factor[key])
                    else:
                        self.extensions_scale_factor[key] = extensions_scale_factor[key]
            # Append to PDF
            self.ln_pdf.append(ln_pdf[:, ln_pdf.nonzero()])
            # Update moment tensor index
            self._i += moment_tensors.shape[1]

    def output(self, normalise=True, convert=False, n_samples=0, discard=10000, mcmc=False):
        """
        Returns nonzero probability samples in a dictionary

        Keyword Arguments
            normalise:[True] boolean flag for normalising the PDF output or not.
            convert:[False] boolean flag for converting the moment tensor output into different parameterisations
            n_samples:[0] Total number of samples (not using self.pdf.n as multiple samplings for e.g. MPI inversions)
            discard:[10000] float, if non-zero discard is a multiplier on n_samples so that samples with probability less than 1/(n_samples*discard) are discarded as negligeable

        Returns
            Dictionary of nonzero probability samples as:
                {'MTSpace':nonzeromoment_tensors,'Probability':nonzeroProbability,'dV':volumeElement}

        Raises
            ValueError: No nonzero probability samples.

        """
        # Gets normalised and unnormalised pdfs
        ln_pdf = self.ln_pdf.output(normalise)
        un_normalised_ln_pdf = self.ln_pdf.output(normalise=False)
        output_string = '\n------------MTfit Forward Model Output------------\n\n'
        # Check if any non-zero samples
        if not np.prod(ln_pdf._ln_pdf.shape):
            output_string += 'No Non-Zero probability samples\n\n'
            return {'probability': []}, output_string
        # Get max probability solutions
        output_string += 'Sample Max Probability: '+str(ln_pdf.max())+'\n'
        moment_tensors = self.moment_tensors[:, :self._i]
        if len(ln_pdf.shape) > 1:
            max_p_mt = unique_columns(moment_tensors[:, np.array(ln_pdf == np.max(ln_pdf._ln_pdf))[0]])
        else:
            max_p_mt = unique_columns(moment_tensors[:, np.array(ln_pdf == np.max(ln_pdf._ln_pdf))])
        output_string += 'Sample Max Probability MT:\n'+str(max_p_mt)+'\n\n'
        # Check if discarding samples
        if discard and n_samples:
            output_string += 'Discarding samples less than 1/'+str(discard*n_samples)+' of the maximum likelihood value as negligeable\n\n'
        # Check if there are samples
        if len(ln_pdf):
            # Get non_zero samples
            non_zero = ln_pdf.nonzero(discard=discard, n_samples=n_samples, _cprob_err=False)
            if discard and n_samples:
                output_string += 'After discard, '+str(non_zero.shape[0])+' samples remain\n\n'
            if len(ln_pdf.shape) > 1:
                probability = np.exp(ln_pdf[:, non_zero])
            else:
                probability = np.exp(ln_pdf[non_zero])
            if len(un_normalised_ln_pdf.shape) > 1:
                un_normalised_ln_pdf = un_normalised_ln_pdf[:, non_zero]
            else:
                un_normalised_ln_pdf = un_normalised_ln_pdf[non_zero]
            output = {'probability': probability, 'dV': ln_pdf.dV, 'ln_pdf': un_normalised_ln_pdf}
            # Multiple events
            if self.number_events > 1:
                if hasattr(self, 'scale_factor'):
                    output['scale_factors'] = list(self.scale_factor[non_zero])
                # Create MTspace for multiple events
                for i in range(self.number_events):
                    output['moment_tensor_space_'+str(i+1)] = moment_tensors[6*i:6*(i+1), non_zero]
                    if convert:
                        # convert results
                        output.update(_convert(output['moment_tensor_space_'+str(i+1)], i+1))
            else:
                output['moment_tensor_space'] = moment_tensors[:, non_zero]
                if convert:
                    # convert results
                    output.update(_convert(output['moment_tensor_space']))
            # Bayesian evidence calculation
            try:
                if not mcmc:
                    output['ln_bayesian_evidence'] = ln_bayesian_evidence(output, n_samples, self._prior)
            except Exception:  # Should fail when McMC and not MC
                pass
            try:
                if not mcmc:
                    # DKL
                    V = 1.0
                    if 'g' not in output.keys():
                        # Multiple events
                        g = sorted([key for key in output.keys() if key == 'g' or ('g' in key and (key[0] == 'g' and key[1] == '_'))], key=lambda u: int(u.split('_')[-1]))
                        d = sorted([key for key in output.keys() if key == 'd' or ('d' in key and (key[0] == 'd' and key[1] == '_'))], key=lambda u: int(u.split('_')[-1]))
                        for i, gi in enumerate(g):
                            if np.max(gi)-np.min(gi) < 0.000001 and np.abs(np.mean(gi)) < 0.000001 and np.max(d[i])-np.min(d[i]) < 0.000001 and np.abs(np.mean(d[i])) < 0.000001:
                                V *= (2*np.pi*np.pi)
                            else:
                                V *= (np.pi*np.pi*np.pi)
                    else:
                        # Single event
                        if np.max(output['g'])-np.min(output['g']) < 0.000001 and np.abs(np.mean(output['g'])) < 0.000001 and np.max(output['d'])-np.min(output['d']) < 0.000001 and np.abs(np.mean(output['d'])) < 0.000001:
                            V *= (2*np.pi*np.pi)
                        else:
                            V *= (np.pi*np.pi*np.pi)
                    output_string += str(V)+'\n'
                    output['dkl'] = dkl_estimate(self.ln_pdf, V, n_samples)
                    output_string += 'PDF Kullback-Leibler Divergence Estimate (Dkl): '+str(output['dkl'])+'\n\n'
            except Exception:
                pass
            if isinstance(output['ln_pdf'], LnPDF):
                output['ln_pdf'] = output['ln_pdf']._ln_pdf
            return output, output_string
        else:
            raise ValueError('No non zero probability samples found.')

    def __len__(self):
        """
        Returns number of tried samples.

        Returns
            n: number of forward modelled MT samples.

        """
        return self.n

    def nonzero(self):
        """Return non zero indices"""
        return self.ln_pdf.nonzero()


class FileSample(Sample):
    """
    FileSample object stores sample values on disk rather than in memory (writes to disk each time)

    These then need to be cleaned up but should allow for simple restore and recover values.

    NEEDS hdf5 (MATLAB -v7.3) storage
    """

    def __init__(self, fname='MTfit_run', file_safe=True, *args, **kwargs):
        """
        FileSample initialisation

        Args
            fname:['MTfit_run'] Filename for storage.
            file_safe:[True] Boolean flag for file safe output (i.e. write and move).

        Returns
            FileSample object

        """
        super(FileSample, self).__init__(*args, **kwargs)
        # Check if hdf5storage is installed
        try:
            from hdf5storage import savemat
            from hdf5storage import loadmat
            self.savemat = savemat
            self.loadmat = loadmat
        except Exception:
            raise ImportError('hdf5storage module missing - required for FileSample sample type')
        # Set default filename
        if not isinstance(fname, str):
            fname = 'MTfit_run'
        self.fname = fname.split('.mat')[0]+'_in_progress.mat'
        self.n = 0
        self._i = 1
        self.non_zero_samples = 0
        self.pdf = LnPDF()
        self.moment_tensor = np.matrix([[]])
        self.file_safe = file_safe
        # Load data if file exists (recover)
        if os.path.exists(self.fname):
            data = self.loadmat(self.fname)
            try:
                self.n = data['n']
                self.non_zero_samples = data['non_zero_samples']
                self._i = data['i']
            except Exception:
                pass

    def append(self, moment_tensors, ln_pdf, n, scale_factor=False, extensions_scale_factor=False):
        """
        Appends new samples to the FileSample file.

        Args
            moment_tensors: numpy matrix of moment tensors to append.
            ln_pdf: numpy matrix/array or PDF object containing ln_pdf samples corresponding to the moment tensors in moment_tensors (must be same length).
            n: number of tried samples (including zero probability samples).
            scale_factor:[False] scale_factor estimates for relative amplitude inversion.
            extensions_scale_factor:[False] scale_factor estimates for extensions carrying out relative inversion.

        Returns
            None

        """
        if isinstance(moment_tensors, list) and isinstance(moment_tensors[0], np.ndarray):
            moment_tensors = np.array(moment_tensors)
            moment_tensors = np.matrix(moment_tensors.reshape(moment_tensors.shape[0]*moment_tensors.shape[1], moment_tensors.shape[2]))
        if isinstance(ln_pdf, (float, long, int, np.float64, np.float32)) and moment_tensors.shape[1] != 1:
            raise ValueError('Moment Tensor shape must be 6x1 when using a float ln_pdf')
        elif not isinstance(ln_pdf, (float, long, int, np.float64, np.float32)) and moment_tensors.shape[1] != ln_pdf.shape[1]:
            raise ValueError('Moment Tensor shape[2] and ln_pdf shape[2] must be the same')

        self.n += n
        if isinstance(ln_pdf, (np.ndarray, float, int, np.float64, np.float32)):
            ln_pdf = LnPDF(ln_pdf)
        moment_tensors = moment_tensors[:, ln_pdf.nonzero()]
        if not isinstance(scale_factor, bool):
            if isinstance(scale_factor, list):
                scale_factor = np.array(scale_factor)
            scale_factor = scale_factor[ln_pdf.nonzero()]
        if isinstance(extensions_scale_factor, dict) and len(extensions_scale_factor):
            # Scale factor dict containing mu and s
            for key in extensions_scale_factor.keys():
                extensions_scale_factor[key] = convert_keys_to_unicode(list(extensions_scale_factor[key][ln_pdf.nonzero()]))
        if self.file_safe:
            old_fname = self.fname
            try:
                shutil.copy(self.fname, self.fname+'~')
            except IOError:
                pass
            self.fname += '~'
        if moment_tensors.shape[0]*moment_tensors.shape[1] > 0:  # Check there are samples
            if len(ln_pdf.nonzero()) != moment_tensors.shape[1]:
                raise ValueError('Moment Tensor shape[2] and ln_pdf shape[2] must be the same')
            self.savemat(self.fname, {'MTSpace_'+str(self._i): moment_tensors}, appendmat=not self.file_safe, store_python_metadata=True)
            if not isinstance(scale_factor, bool):
                # dict with p mu and s
                self.savemat(self.fname, {'scale_factor_'+str(self._i): convert_keys_to_unicode(list(scale_factor))}, appendmat=not self.file_safe, store_python_metadata=True)
            if isinstance(extensions_scale_factor, dict) and len(extensions_scale_factor):
                # dict with p mu and s
                self.savemat(self.fname, {'extensions_scale_factor_'+str(self._i): extensions_scale_factor}, appendmat=not self.file_safe, store_python_metadata=True)
            self.savemat(self.fname, {'LnPDF_'+str(self._i): ln_pdf[:, ln_pdf.nonzero()]}, appendmat=not self.file_safe, store_python_metadata=True)
            self.non_zero_samples += len(ln_pdf.nonzero())
            self.savemat(self.fname, {'non_zero_samples': self.non_zero_samples}, appendmat=not self.file_safe, store_python_metadata=True)
            self._i += 1
        self.savemat(self.fname, {'i': self._i}, appendmat=not self.file_safe, store_python_metadata=True)
        self.savemat(self.fname, {'n': self.n}, appendmat=not self.file_safe, store_python_metadata=True)
        if self.file_safe:
            shutil.copy(self.fname, old_fname)
            self.fname = old_fname

    def output(self, *args, **kwargs):
        """
        Returns nonzero probability samples in a dictionary

        Returns
            Dictionary of nonzero probability samples as:
                {'MTSpace':nonzero_moment_tensors,'Probability':nonzero_probability,'dV':volume_element}

        Raises
            ValueError: No nonzero probability samples.

        """
        # load data
        data = self.loadmat(self.fname)
        # Set up attributes for Sample.output call
        self.moment_tensors = np.matrix(np.zeros((self.number_events*6, self.non_zero_samples)))
        self.scale_factor = False
        if len([key for key in data if 'scale_factor' in key]):
            self.scale_factor = []
        _x = 0
        self.n = data['n']
        for key in [key for key in data.keys() if 'MTSpace' in key]:
            i = key.split('_')[-1]
            self.ln_pdf.append(LnPDF(data['LnPDF_'+str(i)]))
            moment_tensors = data['MTSpace_'+str(i)]
            self.moment_tensors[:, _x: _x+moment_tensors.shape[1]] = moment_tensors
            if isinstance(self.scale_factor, list):
                self.scale_factor.append(data['scale_factor_'+str(i)])
            _x += moment_tensors.shape[1]
        self.scale_factor = np.array(self.scale_factor)
        output, output_string = super(FileSample, self).output(*args, **kwargs)
        try:
            os.remove(self.fname)
        except Exception:
            pass
        return output, output_string


def ln_bayesian_evidence(output, n_samples, prior=_6sphere_prior):
    """
    Bayesian evidence calculation for the output

    The priors are accessible as a pkg_resource, along with the sampling distribution
    (MTfit.algorithms.__base__), so new sampling priors can be added.
    If the sampling is from the prior distribution, the prior in thisfunction is
    just the uniform prior.

    Args
        output: dictionary from Sample.output.
        n_samples: total number of samples test_append.
        prior:['6sphere'] the prior distribution to use (reflects the sampling).

    Returns
        float: returns the Bayesian evidence as a log value.
    """
    # Get prior
    if 'g' not in output.keys():
        # Multiple events
        p = 1.0
        g = sorted([key for key in output.keys() if key == 'g' or ('g' in key and (key[0] == 'g' and key[1] == '_'))], key=lambda u: int(u.split('_')[-1]))
        d = sorted([key for key in output.keys() if key == 'd' or ('d' in key and (key[0] == 'd' and key[1] == '_'))], key=lambda u: int(u.split('_')[-1]))
        for i, gi in enumerate(g):
            p *= prior(output[gi], output[d[i]])
    else:
        # Single event
        p = prior(output['g'], output['d'])
    if not isinstance(output['ln_pdf'], LnPDF):
        output['ln_pdf'] = LnPDF(output['ln_pdf'])
    return np.log((output['ln_pdf']+np.log(p)-output['ln_pdf']._ln_pdf.max()).exp(_cprob_err=False).sum())+output['ln_pdf']._ln_pdf.max()-np.log(n_samples)


def _convert(moment_tensors, i=None):
    """
    Converts the moment tensor to Tape parameters, Hudson u,v and strike dip rake pairs.

    Args
        moment_tensors: moment tensor matrix.
        i:[None] event index if using multiple events.

    Returns
        output dictionary containing converted parameters
    """
    output = output_convert(moment_tensors)
    if isinstance(i, int):
        # Multiple events
        for key in output.keys():
            output[key+'_'+str(i)] = output.pop(key)
    return output



"""
Algorithms
**********

This module contains the sampling approaches used. Two main approaches are used:

 * Random Monte Carlo sampling
 * Markov chain Monte Carlo sampling

However, there are also two variants of the Markov chain Monte Carlo (McMC) method:

 * Metropolis-Hastings
 * Trans-Dimensional Metropolis-Hastings (Reversible Jump)
"""


# **Restricted:  For Non-Commercial Use Only**
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.


from .base import BaseAlgorithm  # noqa E401
from .monte_carlo import IterationSample, TimeSample  # noqa E401
from .markov_chain_monte_carlo import McMCAlgorithmCreator as MarkovChainMonteCarloAlgorithmCreator  # noqa E401


# TODO: refactor sampling logic in these modules and reduce function complexity



"""
file_io
*******

Handles file input/output e.g. for the inversion results
"""


# **Restricted:  For Non-Commercial Use Only**
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.


import os
import sys
import traceback
import struct
from datetime import datetime
import logging

try:
    import cPickle as pickle
except ImportError:
    import pickle

import numpy as np

from ..convert.moment_tensor_conversion import MT6_Tape
from .. import __version__
from ..extensions.scatangle import parse_scatangle


logger = logging.getLogger('MTfit.utilities.file_io')

nlloc_polarity_inv_dict = {'P': {1: 'U', 0: '?', -1: 'D'}, 'S': {1: 'P', 0: '?', -1: 'N'}}
nlloc_polarity_dict = {'u': 1, '?': 0, 'd': -1, '+': 1, 'c': 1, '-': -1, '.': 0, 'p': 1, 'n': -1}

#
# Output Task Classes
#


class MATLABOutputTask(object):

    """
    MATLAB format output task

    Saves MATLAB output file to disk using hdf5storage or scipy savemat.

    Initialisation
        Args
            fid: Filename for MATLAB output.
            output: Dictionary of output to be saved to fid.
    """

    def __init__(self, fid, output_dict, version='7.3'):
        """
        Initialisation of MATLABOutputTask

        Args
            fid: Filename for MATLAB output.
            output: Dictionary of output to be saved to fid.
            version: Set MATLAB version (7.3 or 7) for hdf5storage or scipy storage, respectively

        """
        self.fid = fid
        self.output_dict = output_dict
        self.version = version

    def __call__(self):
        """
        Runs the MATLABOutputTask and returns a result code.

        Returns
            resultCode: 10 if successful, 20 if an exception is thrown.

        """
        # Check version and get savemat function
        if self.version == '7.3':
            try:
                from hdf5storage import savemat
            except Exception:
                self.version = '7'
        if self.version != '7.3':
            from scipy.io import savemat  # noqa F811
            self.version = '7'
        logging.info('Saving to {} in MATLAB version {}'.format(self.fid, self.version))
        # Try to save file
        try:
            # Convert to/from unicode for different functions (hdf5 needs
            # unicode) scipy needs non-unicode
            if self.version == '7.3':
                self.output_dict = convert_keys_to_unicode(self.output_dict)
            else:
                self.output_dict = convert_keys_from_unicode(self.output_dict)
            # Save data
            savemat(self.fid, self.output_dict)
            logger.info('Saved to '+self.fid)
            return 10
        except Exception:
            logger.exception('MATLAB Output Error')
            return 20


class PickleOutputTask(object):

    """
    Pickle format output task

    Saves output file to disk using pickle.

    Initialisation
        Args
            fid: Filename for output.
            output: Dictionary of output to be saved to fid.
    """

    def __init__(self, fid, output_dict,):
        """
        Initialisation of PickleOutputTask

        Args
            fid: Filename for MATLAB output.
            output: Dictionary of output to be saved to fid.

        """
        self.fid = fid
        self.output_dict = output_dict

    def __call__(self):
        """
        Runs the PickleOutputTask and returns a result code.

        Returns
            resultCode: 10 if successful, 20 if an exception is thrown.

        """
        logger.info('Saving to '+self.fid)
        try:
            with open(self.fid, 'wb') as f:
                pickle.dump(self.output_dict, f)
            logger.info('Saved to '+self.fid)
            return 10
        except Exception:
            logger.exception('Pickle Output Error')
            return 20


class HypOutputTask(object):

    """
    NLLOC hyp format output task

    Saves output file to disk.

    Initialisation
        Args
            fid: Filename for output.
            output: Data to be saved to fid.
            mt: Saving mt format.
    """

    def __init__(self, fid, output, binary=False):
        """
        Initialisation of HypOutputTask

        Args
            fid: Filename for MATLAB output.
            output: Data to be saved to fid.
            binary:[False] binary output file.

        """
        self.fid = fid
        self.output = output
        self.binary = binary

    def __call__(self):
        """
        Runs the PickleOutputTask and returns a result code.

        Returns
            resultCode: 10 if successful, 20 if an exception is thrown.

        """
        try:
            # Write binary as binary file,
            if self.binary:
                with open(self.fid, 'wb') as f:
                    f.write(self.output)
            # Write output as normal text
            else:
                with open(self.fid, 'w') as f:
                    f.write(self.output)
            return 10
        except Exception:
            logger.exception('Hyp Output Error')
            return 20

#
# Input file parsing
#


def parse_hyp(filename):
    """
    Parse NonLinLoc hyp file

    Reads NonLinLoc hyp file for input data.

    Args:
        filename:str hyp filename.
    """
    with open(filename) as f:
        lines = f.readlines()
    events_list = []
    event = []
    # Loop over file and split into events
    for line in lines:
        line = line.rstrip()
        line = line.split()
        if len(line):
            event.append(line)
            if line[0] == 'END_NLLOC':
                if len(event):
                    events_list.append(event)
                    event = []
    if len(event):
        events_list.append(event)
    # Parse events
    return _parse_hyp_events(events_list)


def _parse_hyp_events(events_list, polarity_error_multiplier=3.):
    """
    Parse events from NonLinLoc hyp file

    Parse the events read from a NonLinLoc hyp file.

    Args
        events_list: list of event lines read from NonLinLoc hyp file.
        polarity_error_multiplier:[3.] Multiplier to convert time error to polarity error estimates.

    """

    event_dict_list = []
    # Loop over events
    for event in events_list:
        event_dict = {}
        phase = False
        # For each line
        for line in event:
            # Generic key dictionary
            key_dict = {'Stations': {
                'Name': [], 'TakeOffAngle': [], 'Azimuth': []}, 'Measured': [], 'Error': []}
            # Get UID from GEOGRAPHIC line
            if line[0] == 'GEOGRAPHIC':
                sec = line[7]
                try:
                    event_dict['UID'] = line[2]+line[3]+line[4]+line[5]+line[6]+sec.split('.')[0]+sec.split('.')[1][:-1]
                except Exception:
                    pass
            # Phase flag
            if line[0] == 'PHASE':
                phase = True
                continue
            # End of phase flag
            if line[0] == 'END_PHASE':
                phase = False
            # If in phase section, read data
            if phase and len(line) >= 24:
                # Get station
                station = line[0]
                # Get phase
                phase_type = line[4]
                # Get polarity
                polarity = nlloc_polarity_dict[line[5].lower()]
                # Get uncertainty
                uncertainty = float(line[10])
                # Get amplitude (unused)
                amplitude = float(line[12])  # noqa F841
                if polarity != 0:
                    if not event_dict.__contains__(phase_type+'Polarity'):
                        event_dict[phase_type+'Polarity'] = key_dict.copy()
                    event_dict[
                        phase_type+'Polarity']['Stations']['Name'].append(station)
                    event_dict[
                        phase_type+'Polarity']['Measured'].append(polarity)
                    event_dict[
                        phase_type+'Polarity']['Error'].append(uncertainty*polarity_error_multiplier)
                    event_dict[
                        phase_type+'Polarity']['Stations']['TakeOffAngle'].append([float(line[24])])
                    event_dict[
                        phase_type+'Polarity']['Stations']['Azimuth'].append([float(line[23])])
        # End of event, make data into appropriate structures
        for key in event_dict.keys():
            if key != 'UID':
                event_dict[key]['Stations']['TakeOffAngle'] = np.matrix(
                    event_dict[key]['Stations']['TakeOffAngle'])
                event_dict[key]['Stations']['Azimuth'] = np.matrix(
                    event_dict[key]['Stations']['Azimuth'])
                event_dict[key]['Measured'] = np.matrix(
                    event_dict[key]['Measured']).T
                event_dict[key]['Error'] = np.matrix(
                    event_dict[key]['Error']).T
        # Sort keys and append event dict to list
        event_dict['hyp_file'] = event
        if sorted(event_dict.keys()) != ['hyp_file', 'UID']:
            event_dict_list.append(event_dict.copy())
            event_dict = {}
    return event_dict_list


def parse_csv(filename):
    """
    Parses CSV file to data dictionary

    CSV file format is to have events split by blank lines, a header line showing where the information is, UID and data-type information stored in the first column, e.g.

        UID=123,,,,
        PPolarity,,,,
        Name,Azimuth,TakeOffAngle,Measured,Error
        S001,120,70,1,0.01
        S002,160,60,-1,0.02
        P/SHRMSAmplitudeRatio,,,,
        Name,Azimuth,TakeOffAngle,Measured,Error
        S003,110,10,1,0.05 0.04
        ,,,,
        PPolarity ,,,,
        Name,Azimuth,TakeOffAngle,Measured,Error
        S003,110,10,1,0.05

    Is a CSV file with 2 events, one event UID of 123, and PPolarity data at S001 and S002 and P/SHRMSAmplitude data at S003,
    and a second event with no UID (will default to the event number, in this case 2) with PPolarity data at S003.

    Args:
        filename:str csv filename

    Returns:
        event_dict_list:list of event data dictionaries
    """
    # Read file
    with open(filename) as f:
        lines = f.readlines()
    events_list = []
    event = []
    # Loop over lines
    for l in lines:
        if not len(l.split(',')[0]) and not len(l.split(',')[1]):
            # End OF event
            if len(event):
                events_list.append(event)
            event = []
        else:
            event.append(l.rstrip())
    # If there is event data, append to the event list
    if len(event):
        events_list.append(event)
    # Return parsed events
    return _parse_csv_events(events_list)


def _parse_csv_events(events_list):
    """
    Parses CSV event list to data dictionary

    CSV file format is to have events split by blank lines, a header line showing where the information is, UID and data-type information stored in the first column, e.g.

        UID=123,,,,
        PPolarity,,,,
        Name,Azimuth,TakeOffAngle,Measured,Error
        S001,120,70,1,0.01
        S002,160,60,-1,0.02
        P/SHRMSAmplitudeRatio,,,,
        Name,Azimuth,TakeOffAngle,Measured,Error
        S003,110,10,1 2,0.05 0.04
        ,,,,
        PPolarity ,,,,
        Name,Azimuth,TakeOffAngle,Measured,Error
        S003,110,10,1,0.05

    Is a CSV file with 2 events, one event UID of 123, and PPolarity data at S001 and S002 and P/SHRMSAmplitude data at S003,
    and a second event with no UID (will default to the event number, in this case 2) with PPolarity data at S003.

    Args:
        filename:str csv filename

    Returns:
        event_dict_list:list of event data dictionaries
    """
    event_dict_list = []
    # Default keys and values
    key = 'PPolarity'
    name_index = 0
    measured_index = 3
    azimuth_index = 1
    take_off_angle_index = 2
    error_index = 4
    # Loop over events
    for event in events_list:
        event_dict = {'UID': str(events_list.index(event)+1)}
        # Set up generic dictionary
        key_dict = {'Stations': {
            'Name': [], 'TakeOffAngle': [], 'Azimuth': []}, 'Measured': [], 'Error': []}
        # Loop over lines
        for line in event:
            split_line = line.split(',')
            if len(split_line[0]) and not sum([len(u) for u in split_line[1:]]):
                # Try tp get UID
                if 'UID' in split_line[0]:
                    event_dict['UID'] = split_line[0].split(
                        'UID')[1].lstrip(':').lstrip('=').strip()
                else:
                    # End of event
                    if len(key_dict['Stations']['Name']):
                        key_dict['Stations']['TakeOffAngle'] = np.matrix(key_dict['Stations']['TakeOffAngle'])
                        key_dict['Stations']['Azimuth'] = np.matrix(key_dict['Stations']['Azimuth'])
                        key_dict['Measured'] = np.matrix(key_dict['Measured'])
                        key_dict['Error'] = np.matrix(key_dict['Error'])
                        event_dict[key] = key_dict
                    key = split_line[0].strip()
                    key_dict = {'Stations': {
                        'Name': [], 'TakeOffAngle': [], 'Azimuth': []}, 'Measured': [], 'Error': []}
            elif 'Name' in split_line:
                # Try to read header file and get key pieces
                split_line = ','.join(split_line).lower().split(',')
                name_index = split_line.index('name')
                measured_index = split_line.index('measured')
                azimuth_index = split_line.index('azimuth')
                take_off_angle_index = split_line.index('takeoffangle')
                error_index = split_line.index('error')
            else:
                # Append item to list
                key_dict['Stations']['Name'].append(split_line[name_index])
                key_dict['Stations']['TakeOffAngle'].append([float(split_line[take_off_angle_index].strip())])
                key_dict['Stations']['Azimuth'].append([float(split_line[azimuth_index].strip())])
                measured = []
                for _measured in split_line[measured_index].split():
                    measured.append(float(_measured.strip()))
                key_dict['Measured'].append(measured)
                error = []
                for err in split_line[error_index].split():
                    error.append(float(err.strip()))
                key_dict['Error'].append(error)
        # Sort output into correct structs
        if len(key_dict['Stations']['Name']):
            key_dict['Stations']['TakeOffAngle'] = np.matrix(key_dict['Stations']['TakeOffAngle'])
            key_dict['Stations']['Azimuth'] = np.matrix(key_dict['Stations']['Azimuth'])
            key_dict['Measured'] = np.matrix(key_dict['Measured'])
            key_dict['Error'] = np.matrix(key_dict['Error'])
            event_dict[key] = key_dict
        event_dict_list.append(event_dict)
    return event_dict_list

#
# Convert csv to inv
#


def csv2inv(filename):
    """
    Converts CSV file to inv file

    CSV file format is to have events split by blank lines, a header line showing where the information is, UID and data-type information stored in the first column, e.g.::

        UID=123,,,,
        PPolarity,,,,
        Name,Azimuth,TakeOffAngle,Measured,Error
        S001,120,70,1,0.01
        S002,160,60,-1,0.02
        P/SHRMSAmplitudeRatio,,,,
        Name,Azimuth,TakeOffAngle,Measured,Error
        S003,110,10,1,0.05 0.04
        ,,,,
        PPolarity ,,,,
        Name,Azimuth,TakeOffAngle,Measured,Error
        S003,110,10,1,0.05

    Is a CSV file with 2 events, one event UID of 123, and PPolarity data at S001 and S002 and P/SHRMSAmplitude data at S003,
    and a second event with no UID (will default to the event number, in this case 2) with PPolarity data at S003.

    Args:
        filename:str csv filename.

    """
    event_dict_list = parse_csv(filename)
    output_name = os.path.splitext(filename)[0]+'.inv'
    PickleOutputTask(output_name, event_dict_list)()

#
# Hyp output handling
#


def _convert_mt_space_to_struct(output_data, i=False):
    """
    Coverts mt space to binary struct

    Structure is:

        File version
        Nsamples(unsigned long int)
        Nmtspacesamples(unsigned long int)
        Converted(bool)
        Ln_BayesianEvidence
        DKL
        NSamples as:
            P(double)
            Ln_P(double)
            Mnn(double)
            Mee(double)
            Mdd(double)
            Mne(double)
            Mnd(double)
            Med(double)

            if Converted is true then each sample also contains:
                gamma(double)
                delta(double)
                kappa(double)
                h(double)
                sigma(double)
                u(double)
                v(double)
                strike1(double)
                dip1(double)
                rake1(double)
                strike2(double)
                dip2(double)
                rake2(double)

    Args
        output_data: output data dictionaries
        i:[False] event number for multiple events.

    Returns
        binary MT and scale factor outputs
    """
    file_version = 2
    sqrt2 = np.sqrt(2)
    # Check MT is single event
    if 'moment_tensor_space' in output_data:
        # single
        end = ''
    else:
        end = '_'+str(i)
    # Load converted data and set convert flag (for ending)
    try:
        g = output_data['g'+end]
        d = output_data['d'+end]
        k = output_data['k'+end]
        h = output_data['h'+end]
        s = output_data['s'+end]
        u = output_data['u'+end]
        v = output_data['v'+end]
        s1 = output_data['S1'+end]
        d1 = output_data['D1'+end]
        r1 = output_data['R1'+end]
        s2 = output_data['S2'+end]
        d2 = output_data['D2'+end]
        r2 = output_data['R2'+end]
        converted = True
    except Exception:
        converted = False
    # Add Number of samples and converted flag
    binary_output = struct.pack('QQQ?', file_version, output_data['total_number_samples'],
                                output_data['moment_tensor_space'+end].shape[1], converted)
    # Try to add Bayesian evidence
    try:
        binary_output += struct.pack('1d', output_data['ln_bayesian_evidence'])
    except Exception:
        binary_output += struct.pack('1d', np.nan)  # No Bayesian evidence
    # Try to add Dkl
    try:
        binary_output += struct.pack('1d', output_data['dkl'])
    except Exception:
        binary_output += struct.pack('1d', np.nan)  # No Dkl
    # Loop over MTs
    for i in range(output_data['moment_tensor_space'+end].shape[1]):
        # This seems slow in python 3
        # Add MT data
        if output_data['probability'].ndim == 2:
            p = output_data['probability'][0, i]
        else:
            p = output_data['probability'][i]
        if output_data['ln_pdf'].ndim == 2:
            ln_p = output_data['ln_pdf'][0, i]
        else:
            ln_p = output_data['ln_pdf'][i]
        binary_output += struct.pack('8d', p, ln_p, output_data['moment_tensor_space'+end][0, i], output_data['moment_tensor_space'+end][1, i],
                                     output_data['moment_tensor_space'+end][2, i], output_data['moment_tensor_space'+end][3, i]/sqrt2,
                                     output_data['moment_tensor_space'+end][4, i]/sqrt2,
                                     output_data['moment_tensor_space'+end][5, i]/sqrt2)
        if converted:
            # Add converted data
            binary_output += struct.pack('13d', g[i], d[i], k[i], h[i], s[i], u[i], v[i], s1[i], d1[i], r1[i], s2[i], d2[i], r2[i])
    # Handle scale_factors
    if sys.version_info.major > 2:
        sf_output = b''
    else:
        sf_output = ''
    if 'scale_factors' in output_data:
        n_events = output_data['scale_factors']['mu'].shape[1]
        sf_output = struct.pack('QQQ', output_data['total_number_samples'], len(output_data['scale_factors']), n_events)
        for i, sf in enumerate(output_data['scale_factors']):
            sf_output += struct.pack('dd', output_data['probability'][0, i], output_data['ln_pdf'][0, i])
            # Add data for each event - Loops over off diagonal elements
            k = 0
            _l = 1
            for j in range(int(n_events*(n_events-1)/2.)):
                sf_output += struct.pack('dd', sf['mu'][k, _l], sf['sigma'][k, _l])
                _l += 1
                if _l >= n_events:
                    k += 1
                    _l = k+1
    # Return output
    return binary_output, sf_output


def _generate_hyp_output_data(event_data, inversion_options=False, output_data=False, maxMT=np.matrix([[0], [0], [0], [0], [0], [0]])):
    """
    Generates the hyp output text data for the .hyp file

    Appends or creates the hyp output file (if a hyp file is used as input, the text is stored in the event data)

    Args
        event_data: Event data dictionary.
        inversion_options:[False] Inversion options list.
        output_data:[False] Output data results
        maxMT:[np.matrix([[0],[0],[0],[0],[0],[0]])] Maximum moment tensor solution.

    Returns
        list of event text outputs
    """
    # Try to get the maximum moment tensor
    from ..inversion import _polarity_misfit_check
    maxMT_tape = [0, 0, 0, 0, 0]
    try:
        maxMT = unique_columns(output_data['moment_tensor_space'][:, np.array(
            output_data['probability'] == output_data['probability'].max())[0]])
        if maxMT.shape[1] > 1:
            maxMT = maxMT[:, 0]
        maxMT_tape = MT6_Tape(maxMT)
    except Exception:
        traceback.print_exc()
    if all(maxMT == 0):
        maxMT_tape = [0, 0, 0, 0, 0]
    # Check if its a DC
    dc = False
    if abs(maxMT_tape[0]) < 1.0*10**-5 and abs(maxMT_tape[1]) < 1.0*10**-5:
        dc = True
    # Try to get the input hyp file if it exists
    if 'hyp_file' in event_data.keys():
        lines = event_data['hyp_file'][:]
    else:
        # Generate NLLOC type hyp file
        lines = []
        try:
            lines.append(['NLLOC', event_data['UID']])
        except Exception:
            lines.append(['NLLOC'])
        lines.append(['SIGNATURE', 'MTfit', __version__+'/'+datetime.now().isoformat()])
        lines.append(['COMMENT', 'MTfit inversion'])
        lines.append(['GRID', 'None'])
        lines.append(['SEARCH', 'None'])
        lines.append(['HYPOCENTER', 'x', '?', 'y', '?', 'z', '?',
                      'OT', '?', 'ix', '?', 'iy', '?', 'iz', '?'])
        lines.append(['GEOGRAPHIC', 'OT', '?', '?', '?', '?',
                      '?', '?', 'Lat', '?', 'Long', '?', 'Depth', '?'])
        lines.append(['QUALITY', 'Pmax', '?', 'MFmin', '?', 'MFmax', '?', 'RMS', '?',
                      'Nphs', '?', 'Gap', '?', 'Dist', '?', 'Mamp', '?', '?', 'Mdur', '?', '?'])
        lines.append(['VPVSRATIO', '?', '?'])
        lines.append(['STATISTICS', 'ExpectX', '?', 'Y', '?', 'Z', '?', 'CovXX', '?', 'XY', '?', 'XZ', '?', 'YY', '?',
                      'YZ', '?', 'ZZ', '?', 'EllAz1', '?', 'Dip1', '?', 'Len1', '?', 'Az2', '?', 'Dip2', '?', 'Len2', '?', 'Len3', '?'])
        lines.append(['TRANS', 'None'])
        lines.append(['PHASE', 'ID', 'Ins', 'Cmp', 'On', 'Pha', 'FM', 'Date', 'HrMn', 'Sec', 'Err', 'ErrMag', 'Coda', 'Amp',
                      'Per', '>', 'TTpred', 'Res', 'Weight', 'StaLoc(X', 'Y', 'Z)', 'SDist', 'SAzim', 'RAz', 'RDip', 'RQual', 'Tcorr'])
        # Station Data
        for phase in [key.split('Polarity')[0] for key in event_data.keys() if 'Polarity' in key and 'PolarityProb' not in key]:
            for i, station in enumerate(event_data[phase+'Polarity']['Stations']['Name']):
                polarity = nlloc_polarity_inv_dict[phase[0]][
                    event_data[phase+'Polarity']['Measured'][i, 0]]
                lines.append([station, '?', '?', '?', phase.upper(), polarity, '?', '?', '?', '?', '?', '?', '?', '?', '>', '?', '?', '?', '?', '?', '?', '?', '?',
                              '{:5.1f}'.format(event_data[phase+'Polarity']['Stations']['Azimuth'][i, 0]).lstrip(),
                              '{:5.1f}'.format(event_data[phase+'Polarity']['Stations']['TakeOffAngle'][i, 0]).lstrip(), '?', '?'])
        lines.append(['END_PHASE'])
        lines.append(['END_NLLOC'])
    phase_line_index = lines.index(['PHASE', 'ID', 'Ins', 'Cmp', 'On', 'Pha', 'FM', 'Date', 'HrMn', 'Sec', 'Err', 'ErrMag', 'Coda',
                                    'Amp', 'Per', '>', 'TTpred', 'Res', 'Weight', 'StaLoc(X', 'Y', 'Z)', 'SDist', 'SAzim', 'RAz', 'RDip', 'RQual', 'Tcorr'])
    end_phase_line_index = lines.index(['END_PHASE'])
    geog_line = [line for line in lines if line[0] == 'GEOGRAPHIC'][0]
    # GET PHASE LINE AND THEN PREPEND MT AND UPDATE misfit_checks
    # DC  FOCALMECH dlat dlong depth Mech dipDir dipAng rake mf misfit nObs numberObs
    # MT  MOMENTTENSOR dlat dlong depth MTNN Mnn EE Mee DD Mdd NE Mne ND Mnd
    # ED Med mf misfit nObs numberObs
    try:
        source_line_index = lines.index(
            [line for line in lines if line[0] == 'FOCALMECH'][0])
        lines.pop(source_line_index)
    except Exception:
        pass
    # Make source line
    if dc:
        source_line = ["FOCALMECH", geog_line[9], geog_line[11], geog_line[13], "Mech", str(float(maxMT_tape[2]*180/np.pi)),
                       str(float(np.arccos(maxMT_tape[3])*180/np.pi)), str(float(maxMT_tape[4]*180/np.pi)), "mf", "?", "nObs", "?"]
    else:
        source_line = ["MOMENTTENSOR", geog_line[9], geog_line[11], geog_line[13], "MTNN", str(float(maxMT[0])), "EE", str(float(maxMT[1])),
                       "DD", str(float(maxMT[2])), "NE", str(float(maxMT[3]/np.sqrt(2))), "ND", str(float(maxMT[4]/np.sqrt(2))), "ED",
                       str(float(maxMT[5]/np.sqrt(2))), "mf", "?", "nObs", "?"]
    # Insert before phases start
    lines.insert(phase_line_index, source_line)
    mf = 0
    nobs = 0
    # Run polarity misfit checks
    for i in range(phase_line_index+2, end_phase_line_index+1):
        if len(lines[i]) < 24:
            continue
        if _polarity_misfit_check(nlloc_polarity_dict[lines[i][5].lower()], float(lines[i][23]), float(lines[i][24]), lines[i][4], maxMT):
            mf += 1
            lines[i][5] = nlloc_polarity_inv_dict[lines[i][4].upper()[0]][nlloc_polarity_dict[lines[i][5].lower()]].lower()
        else:
            lines[i][5] = nlloc_polarity_inv_dict[lines[i][4].upper()[0]][nlloc_polarity_dict[lines[i][5].lower()]].upper()

        if lines[i][5] != '?':
            nobs += 1
    # Set number of misfits and number of observations
    lines[phase_line_index][-3] = str(mf)
    lines[phase_line_index][-1] = str(nobs)
    output = []
    for line in lines:
        output.append(' '.join(line))
    return output

#
# Results data format
#


def full_pdf_output_dicts(event_data, inversion_options=False, output_data=False, location_samples=False,
                          location_sample_multipliers=False, multiple_events=False, _diagnostic_output=False,
                          normalise=True, *args, **kwargs):
    """
    Create output dictionaries for full_pdf format

    This creates the output dictionaries for the full_pdf format from the output data.

    Args
        event_data: Event data containing stations and observations (input data for the inversion).
        inversion_options:[False] List of inversion data type options.
        output_data:[False] Output data dictionaries (output from algorithm.__output__()).
        location_samples:[False] List of location PDF samples.
        location_sample_multipliers:[False] List of location PDF sample probabilities.
        multiple_events:[False] Boolean flag for multiple events.
        _diagnostic_output:[False] Output angle coefficents and other data for testing and diagnostics.
        normalise:[True] Normalise the output probabilities.

    Keyword Arguments
        station_only:[False] Boolean flag to only output station data (no inversion data).

    Returns
        [mdict,sdict]: list of results dict and station distribution dict.
    """
    # Check if station only output (in kwargs)
    station_output = kwargs.get('station_only', False)
    # Set up output dicts
    stations = {}
    other = {}
    station_distribution = {}
    # Get number of events:
    n_events = 1
    if multiple_events:
        n_events = len(event_data)
    all_stations = []
    # Loop over event_data
    for i in range(n_events):
        if isinstance(event_data, list):
            ev_data = event_data[i]
        else:
            ev_data = event_data
        # If P Polarity or P Polarity probability data used, then get
        # polarities from that
        if (inversion_options and 'PPolarity' in inversion_options and 'PPolarity' in ev_data.keys()) \
                or ('PPolarity' in ev_data.keys() and len(ev_data['PPolarity']['Stations']['Name'])) \
                or (inversion_options and 'PPolarityProbability' in inversion_options and 'PPolarityProbability' in ev_data.keys()):
            if (inversion_options and 'PPolarityProbability' in inversion_options and 'PPolarityProbability' in ev_data.keys()):
                observations_dict = ev_data['PPolarityProbability']
                if observations_dict['Measured'].shape[1] == 2:
                    positive = np.array(
                        observations_dict['Measured'][:, 0]).flatten()
                    negative = np.array(
                        observations_dict['Measured'][:, 1]).flatten()
                else:
                    positive = np.array(
                        observations_dict['Measured'][0, :].T).flatten()
                    negative = np.array(
                        observations_dict['Measured'][1, :].T).flatten()
                measured = positive
                measured[positive < negative] = -negative[positive < negative]
                measured = np.matrix(measured).T
            else:
                observations_dict = ev_data['PPolarity']
                if len(observations_dict['Measured'].shape) > 1 and observations_dict['Measured'].shape[0] > observations_dict['Measured'].shape[1]:
                    measured = np.matrix(observations_dict['Measured'])
                else:
                    measured = np.matrix(observations_dict['Measured']).T
            number_stations = len(observations_dict['Stations']['Name'])
            # Make stations object
            stations = np.matrix(np.zeros((number_stations, 4), dtype=np.object))
            stations[:, 0] = np.matrix(observations_dict['Stations']['Name']).T
            # Check orientations of angles and observations
            if len(observations_dict['Stations']['Azimuth'].shape) > 1 and observations_dict['Stations']['Azimuth'].shape[0] > observations_dict['Stations']['Azimuth'].shape[1]:
                stations[:, 1] = np.matrix(observations_dict['Stations']['Azimuth'])
            else:
                stations[:, 1] = np.matrix(observations_dict['Stations']['Azimuth']).T
            if len(observations_dict['Stations']['TakeOffAngle'].shape) > 1 and observations_dict['Stations']['TakeOffAngle'].shape[0] > observations_dict['Stations']['TakeOffAngle'].shape[1]:
                stations[:, 2] = np.matrix(observations_dict['Stations']['TakeOffAngle'])
            else:
                stations[:, 2] = np.matrix(observations_dict['Stations']['TakeOffAngle']).T
            stations[:, 3] = measured
            # Append to array
            stations = np.array(stations)
        # Otherwise just get set of stations in data (loop over observations
        # dictionaries).
        else:
            # Get all observation dicts used in the inversion
            if inversion_options and inversion_options != 'False':
                observations_dicts = [
                    ev_data[method] for method in inversion_options if method in ev_data.keys()]
            # Otherwise get all data
            else:
                observations_dicts = [
                    val for key, val in ev_data.items() if key not in ['hyp_file', 'UID']]
            stations = np.matrix(np.zeros((0, 4), dtype=np.object))
            for j, observations_dict in enumerate(observations_dicts):
                try:
                    number_stations = len(observations_dict['Stations']['Name'])
                except Exception:
                    continue
                _stations = np.matrix(np.zeros((number_stations, 4), dtype=np.object))
                _stations[:, 0] = np.matrix(observations_dict['Stations']['Name']).T
                observations_dict['Stations']['Azimuth'] = np.matrix(observations_dict['Stations']['Azimuth'])
                if observations_dict['Stations']['Azimuth'].shape[0] < observations_dict['Stations']['Azimuth'].shape[1]:
                    observations_dict['Stations']['Azimuth'] = observations_dict['Stations']['Azimuth'].T
                observations_dict['Stations']['TakeOffAngle'] = np.matrix(observations_dict['Stations']['TakeOffAngle'])
                if observations_dict['Stations']['TakeOffAngle'].shape[0] < observations_dict['Stations']['TakeOffAngle'].shape[1]:
                    observations_dict['Stations']['TakeOffAngle'] = observations_dict['Stations']['TakeOffAngle'].T
                _stations[:, 1] = np.matrix(observations_dict['Stations']['Azimuth'])
                _stations[:, 2] = np.matrix(observations_dict['Stations']['TakeOffAngle'])
                _stations[:, 3] = np.zeros((len(observations_dict['Stations']['Name']), 1))
                try:
                    stations = np.append(stations, _stations, 0)
                except Exception:
                    pass
                stations = np.array(stations)
            # Remove duplicate stations
            if len(stations):
                sta, ind = np.unique(np.array(stations[:, 0].tolist()), True)
                stations = stations[ind, :]
        if n_events > 1:
            all_stations.append(stations)
        else:
            all_stations = stations
    # Station distributions
    if location_samples and location_sample_multipliers:
        station_distribution['Distribution'] = location_samples
        station_distribution['Probability'] = location_sample_multipliers
    # Add inversion options
    other['Inversions'] = inversion_options
    # Add diagnostic options if possible
    if _diagnostic_output:
        a_polarity = kwargs.get('a_polarity', False)
        other['a_polarity'] = a_polarity
        error_polarity = kwargs.get('error_polarity', False)
        other['error_polarity'] = error_polarity
        a1_amplitude_ratio = kwargs.get('a1_amplitude_ratio', False)
        other['a1_amplitude_ratio'] = a1_amplitude_ratio
        percentage_error1_amplitude_ratio = kwargs.get('percentage_error1_amplitude_ratio', False)
        other['percentage_error1_amplitude_ratio'] = percentage_error1_amplitude_ratio
        a2_amplitude_ratio = kwargs.get('a2_amplitude_ratio', False)
        other['a2_amplitude_ratio'] = a2_amplitude_ratio
        percentage_error2_amplitude_ratio = kwargs.get('percentage_error2_amplitude_ratio', False)
        other['percentage_error2_amplitude_ratio'] = percentage_error2_amplitude_ratio
    # Construct event dict
    event = {'NSamples': output_data.pop('total_number_samples'), 'dV': output_data.pop('dV'), 'Probability': output_data.pop('probability')}
    try:
        event['UID'] = event_data['UID']
    except Exception:
        pass
    output_data_keys = list(output_data.keys())
    # Handle multiple events data types
    if not station_output:
        for key in output_data_keys:
            if key == 'moment_tensor_space':
                event['MTSpace'] = output_data.pop('moment_tensor_space')
            elif 'moment_tensor_space' in key:
                event['MTSpace'+key.split('_')[-1]] = output_data.pop(key)
            if key in ['g', 'd', 'h', 'k', 's', 'u', 'v', 'S1', 'D1', 'R1', 'S2', 'R2', 'D2']:
                event[key] = output_data.pop(key)
            elif key.split('_')[0] in ['g', 'd', 'h', 'k', 's', 'u', 'v', 'S1', 'D1', 'R1', 'S2', 'R2', 'D2']:
                event[key] = output_data.pop(key)
        event.update(output_data)
    else:
        event = {'NSamples': 0}
    try:
        event['ln_pdf'] = event['ln_pdf']._ln_pdf
    except Exception:
        pass
    # Create mdict and sdict and return
    if (isinstance(all_stations, list) and not len(all_stations)) or (isinstance(all_stations, np.ndarray) and not np.prod(all_stations.shape)):
        all_stations = []
    mdict = {'Events': event, 'Stations': all_stations, 'Other': other}
    sdict = {'StationDistribution': station_distribution}
    if not len(sdict['StationDistribution']):
        sdict = False
    return [mdict, sdict]


def hyp_output_dicts(event_data, inversion_options=False, output_data=False, location_samples=False, location_sample_multipliers=False, multiple_events=False, _diagnostic_output=False, normalise=True, *args, **kwargs):
    """
    Create output dictionaries for hyp format

    This creates the output dictionaries for the hyp format from the output data.

    Args
        event_data: Event data containing stations and observations (input data for the inversion).
        inversion_options:[False] List of inversion data type options.
        output_data:[False] Output data dictionaries (output from algorithm.__output__()).
        location_samples:[False] List of location PDF samples.
        location_sample_multipliers:[False] List of location PDF sample probabilities.
        multiple_events:[False] Boolean flag for multiple events.
        _diagnostic_output:[False] Output angle coefficents and other data for testing and diagnostics.
        normalise:[True] Normalise the output probabilities.

    Keyword Arguments
        station_only:[False] Boolean flag to only output station data (no inversion data).

    Returns
        [output_data,output_mt_binary,output_scale_factor_binary]: list of results as text and binary mt and scale_factor structures.
    """
    # Check Multiple events
    if multiple_events:
        output_contents = []
        output_mt = []
        output_sf = []
        # Loop over events
        for i, ev_data in enumerate(event_data):
            if 'moment_tensor_space' not in output_data:
                try:
                    ind = np.array(output_data['probability'] == output_data['probability'].max())[0]
                    maxMT = unique_columns(output_data['moment_tensor_space_'+str(i+1)][:, ind])
                    if maxMT.shape[1] > 1:
                        maxMT = maxMT[:, 0]
                except Exception:
                    maxMT = np.matrix([[0], [0], [0], [0], [0], [0]])
            # Get hyp output data
            output_contents.append('\n'.join(_generate_hyp_output_data(ev_data, inversion_options, output_data, maxMT))+'\n\n')
            # Get mt and scale_factor outputs
            binary, sf = _convert_mt_space_to_struct(output_data, i+1)
            output_mt.append(binary)
            output_sf.append(sf)
    else:
        # Get hyp output data
        output_contents = ['\n'.join(_generate_hyp_output_data(event_data, inversion_options, output_data))+'\n']
        # Get mt and scale_factor outputs
        binary, sf = _convert_mt_space_to_struct(output_data)
        # Need to convert this to a string if using python 3
        output_mt = [binary]
        output_sf = [sf]
    if sys.version_info.major > 2:
        binary_spacer = b''
    else:
        binary_spacer = ''
    return '\n'.join(output_contents), binary_spacer.join(output_mt), binary_spacer.join(output_sf)

#
# Output file formats
#


def MATLAB_output(output_data, fid='MTfitOutput.mat', pool=False, version='7.3', *args, **kwargs):
    """
    Outputs event results to fid as .mat

    HDF5 is used as default (MATLAB -v7.3 file types), but this can be set using the version keyword.
    If hdf5storage is not installed then it defaults to pre version 7.3 - can lead to filesize problems with large files.

    Saves MATLAB output dictionary to fid

    Args
        output_data: data dictionary for the event_data.
        fid:['MTfitOutput.mat'] Filename for output.
        pool:[False] Use Jobpool for parallel output.
        version:[7.3] MATLAB file type to use (v 7.3 requires hdf5storage and h5py).

    Returns
        output_string, fid: String of information for stdout, filename of output file.

    """
    # Check MATLAB file version (7.3 requires hdf5storage and h5py but can
    # store files >2Gb)
    if version == '7.3':
        try:
            from hdf5storage import savemat
        except ImportError:
            from scipy.io import savemat
            version = '7'
    else:
        from scipy.io import savemat  # noqa F401
    try:
        from scipy.io import savemat as st_savemat
        st_ver = '7'
    except ImportError:
        from hdf5storage import savemat as st_savemat  # noqa F401
        st_ver = '7.3'
    output_string = 'Outputting data in MATLAB format --version={}\n'.format(version)
    # Get sdict and mdict
    sdict = False
    if isinstance(output_data, list):
        mdict = output_data[0]
        if len(output_data) > 1:
            sdict = output_data[1]
    else:
        mdict = output_data
    # Convert dict keys to or from unicode
    if version == '7.3':
        mdict = convert_keys_to_unicode(mdict)
    else:
        mdict = convert_keys_from_unicode(mdict)
    if st_ver == '7.3' and sdict:
        sdict = convert_keys_to_unicode(sdict)
    else:
        sdict = convert_keys_from_unicode(sdict)
    # Run output tasks (using pool or otherwise)
    if pool and (sys.version_info[:2] >= (2, 7, 4) or ('Events' in mdict.keys() and np.sum([np.prod(mdict['Events'][key].shape) for key in mdict['Events'] if 'MTSpace' in key])*8*8 < 2**30)):
        # Check for cPickle bug #13555 http://bugs.python.org/issue13555 which seems linked to multiprocessing issue #17560 http://bugs.python.org/issue17560
        # cannot pickle files longer than 2**31 (32 bit encoding used for
        # cPickle length)
        output_string += 'Using jobPool to save file to {}\n'.format(fid)
        if mdict:
            pool.custom_task(MATLABOutputTask, os.path.splitext(fid)[0]+'.mat', mdict, version)
        if sdict:
            pool.custom_task(MATLABOutputTask, os.path.splitext(fid)[0]+'StationDistribution.mat', sdict, st_ver)
    else:
        if mdict:
            MATLABOutputTask(os.path.splitext(fid)[0]+'.mat', mdict, version)()
        if sdict:
            try:
                MATLABOutputTask(os.path.splitext(fid)[0]+'StationDistribution.mat', sdict, st_ver)()
            except Exception:
                traceback.print_exc()
        output_string += 'Saved to {}\n'.format(fid)
    return output_string, fid


def pickle_output(output_data, fid='MTfitOutput.out', pool=False, *args, **kwargs):
    """
    Outputs event results to fid as .out (default)

    cPickle output of MATLAB format output.

    Saves output dictionary to fid

    Args
        output_data: data dictionary for the event_data.
        fid:['MTfitOutput.out'] Filename for output.
        pool:[False] Use Jobpool for parallel output.

    Returns
        output_string,fid: String of information for stdout, filename of output file.
    """
    # Set file ending
    if os.path.splitext(fid)[1] == '.mat':
        fid = os.path.splitext(fid)[0]+'.out'
    sdict = False
    if isinstance(output_data, list):
        mdict = output_data[0]
        if len(output_data) > 1:
            sdict = output_data[1]
    else:
        mdict = output_data
    output_string = 'Outputting data in python format\n'
    # Output structure to cPickle.
    if pool and (sys.version_info[:2] >= (2, 7, 4) or ('Events' in mdict.keys() and np.sum([np.prod(mdict['Events'][key].shape) for key in mdict['Events'] if 'MTSpace' in key])*8*8 < 2**30)):
        # Check for cPickle bug #13555 http://bugs.python.org/issue13555 which seems linked to multiprocessing issue #17560 http://bugs.python.org/issue17560
        # cannot pickle files longer than 2**31 (32 bit encoding used for
        # cPickle length)
        output_string += 'Using jobPool to save file to '+fid+'\n'
        pool.custom_task(PickleOutputTask, fid, mdict)
        if sdict:
            pool.custom_task(PickleOutputTask, os.path.splitext(fid)[0]+'StationDistribution'+os.path.splitext(fid)[1], sdict)
    else:
        PickleOutputTask(fid, mdict)()
        if sdict:
            PickleOutputTask(os.path.splitext(fid)[0]+'StationDistribution'+os.path.splitext(fid)[1], sdict)()
        output_string += 'Saved to '+fid+'\n'
    return output_string, fid


def hyp_output(output_data, fid='MTfitOutput.hyp', pool=False, *args, **kwargs):
    """
    Outputs event results to fid as .hyp (default)

    hyp format output.

    Saves output dictionary to fid

    Args
        output_data: data dictionary for the event_data.
        fid:['MTfitOutput.hyp'] Filename for output.
        pool:[False] Use Jobpool for parallel output.

    Returns
        output_string,fid: String of information for stdout, filename of output file.
    """
    # Set file ending
    if os.path.splitext(fid)[1] == '.mat':
        fid = os.path.splitext(fid)[0]+'.hyp'
    mt_data = False
    sf_data = False
    # Get data
    if isinstance(output_data, (list, tuple)):
        output = output_data[0]
        if len(output_data) > 1:
            mt_data = output_data[1]
        if len(output_data) > 2:
            sf_data = output_data[2]
    else:
        output = output_data
    output_string = 'Outputting data in NLLOC hyp format\n'
    # Output using pool or otherwise
    if pool:
        # Check for cPickle bug #13555 http://bugs.python.org/issue13555 which seems linked to multiprocessing issue #17560 http://bugs.python.org/issue17560
        # cannot pickle files longer than 2**31 (32 bit encoding used for
        # cPickle length)
        output_string += 'Using jobPool to save file to '+fid+'\n'
        pool.custom_task(HypOutputTask, fid, output, False)
        if mt_data and (sys.version_info[:2] >= (2, 7, 4) or len(mt_data)*128 < 2**30):
            pool.custom_task(
                HypOutputTask, fid.replace('.hyp', '.mt'), mt_data, True)
        if sf_data and (sys.version_info[:2] >= (2, 7, 4) or len(sf_data)*128 < 2**30):
            pool.custom_task(
                HypOutputTask, fid.replace('.hyp', '.sf'), sf_data, True)

    else:
        HypOutputTask(fid, output, False)()
        if mt_data:
            HypOutputTask(fid.replace('.hyp', '.mt'), mt_data, True)()
        if sf_data and len(sf_data):
            HypOutputTask(fid.replace('.hyp', '.sf'), sf_data, True)()
        output_string += 'Saved to '+fid+'\n'
    return output_string, fid

#
# Read inversion output
#


def read_binary_output(filename, version=2):
    """
    Reads binary output and converts to data dict.

    Reads the binary output created using hyp output format (_convert_mt_space_to_struct) and returns an output data dictionary.

    Args
        filename: str filename to read.

    Keyword Args
        version: int version of the file format (Outputs from versions before 1.0.3 did not contain
                 the DKL estimate or file version, so the version must be set to 1 for these,
                 otherwise the version is set automatically).

    Returns
        output_data: list of dictionaries of output data.
    """
    # Open file
    with open(filename, 'rb') as f:
        f.seek(0, 2)
        nmax = f.tell()
        f.seek(0, 0)
        output_data = []
        sqrt2 = np.sqrt(2)
        while f.tell() < nmax:
            # Read file version
            if version >= 2:
                version = struct.unpack('Q', f.read(8))[0]
            # Read total number and number of saved samples and converted flag
            total_number_samples, number_mt_samples, converted = struct.unpack('QQ?', f.read(17))
            # Get the Bayesian Evidence estimate
            ln_bayesian_evidence = struct.unpack('1d', f.read(8))[0]
            # Get the DKL estimate
            if version >= 2:
                dkl = struct.unpack('1d', f.read(8))[0]
            else:
                dkl = np.nan
            # Generate blank arrays
            MTSpace = np.matrix(np.zeros((6, number_mt_samples)))
            Probability = np.matrix(np.zeros((1, number_mt_samples)))
            Ln_P = np.matrix(np.zeros((1, number_mt_samples)))
            if converted:
                g = np.zeros((number_mt_samples,))
                d = np.zeros((number_mt_samples,))
                k = np.zeros((number_mt_samples,))
                h = np.zeros((number_mt_samples,))
                s = np.zeros((number_mt_samples,))
                u = np.zeros((number_mt_samples,))
                v = np.zeros((number_mt_samples,))
                s1 = np.zeros((number_mt_samples,))
                d1 = np.zeros((number_mt_samples,))
                r1 = np.zeros((number_mt_samples,))
                s2 = np.zeros((number_mt_samples,))
                d2 = np.zeros((number_mt_samples,))
                r2 = np.zeros((number_mt_samples,))
            # Loop over samples
            for i in range(number_mt_samples):
                Probability[0, i], Ln_P[0, i], Mnn, Mee, Mdd, Mne, Mnd, Med = struct.unpack('8d', f.read(64))
                MTSpace[0, i] = Mnn
                MTSpace[1, i] = Mee
                MTSpace[2, i] = Mdd
                MTSpace[3, i] = sqrt2*Mne
                MTSpace[4, i] = sqrt2*Mnd
                MTSpace[5, i] = sqrt2*Med
                if converted:
                    g[i], d[i], k[i], h[i], s[i], u[i], v[i], s1[i], d1[i], r1[
                        i], s2[i], d2[i], r2[i] = struct.unpack('13d', f.read(104))
            # create output data dict
            out = {'moment_tensor_space': MTSpace, 'dkl': dkl, 'probability':
                   Probability, 'ln_pdf': Ln_P, 'total_number_samples': total_number_samples}
            if converted:
                out.update({'g': g, 'd': d, 'k': k, 'h': h, 's': s, 'u': u, 'v': v, 'S1': s1, 'D1': d1,
                            'R1': r1, 'S2': s2, 'D2': d2, 'R2': r2, 'ln_bayesian_evidence': ln_bayesian_evidence})
            # Append to output data list
            output_data.append(out)
    return output_data


def read_sf_output(filename):
    """
    Reads binary scale factor output and converts to data dict.

    Reads the binary scalefactor output created using hyp output format (_convert_mt_space_to_struct) and returns an output data dictionary.
    The scalefactor pairs are stored as an array with the indices corresponding to the pairs 0,1 0,2 0,3 ... 1,2 1,3 .... etc.

    Args
        filename: str filename to read.

    Returns
        output_data: list of dictionaries of output data.
    """
    # Open file
    with open(filename, 'rb') as f:
        f.seek(0, 2)
        nmax = f.tell()
        f.seek(0, 0)
        output_data = []
        while f.tell() < nmax:
            # Read total number and number of saved samples and number of events
            total_number_samples, number_samples, n_events = struct.unpack('QQQ', f.read(24))
            # Generate blank arrays
            Mu = np.matrix(np.zeros((int(n_events*(n_events-1)/2.), number_samples)))
            S = np.matrix(np.zeros((int(n_events*(n_events-1)/2.), number_samples)))
            Ln_P = np.matrix(np.zeros((1, number_samples)))
            Probability = np.matrix(np.zeros((1, number_samples)))
            # Loop over samples
            for i in range(number_samples):
                Probability[0, i], Ln_P[0, i] = struct.unpack('2d', f.read(16))
                # Loop over event pairs (stored as 0,1 0,2 0,3 ... 1,2 1,3 ....)
                for j in range(int(n_events*(n_events-1)/2.)):
                    Mu[j, i], S[j, i] = struct.unpack('dd', f.read(16))
            # Creat output dict
            out = {'scale_factors': Mu, 'probability': Probability, 'ln_pdf':
                   Ln_P, 'total_number_samples': total_number_samples, 'sigma': S}
            # Append to output list
            output_data.append(out)
    return output_data


def read_matlab_output(filename, station_distribution=False):
    """
    Reads matlab output and converts to data dict.

    Reads the matlab output created using the matlab output format and returns an output data dictionary.

    Args
        filename: str filename to read.

    Returns
        (dict,dict): tuple of (events,data) dictionaries of output data.
    """
    try:
        hdf5 = False
        from scipy.io import loadmat
        data = loadmat(filename)
    except Exception:
        try:
            from hdf5storage import loadmat
            hdf5 = True
            data = loadmat(filename)
        except Exception:
            if hdf5:
                raise IOError('Input file '+filename+" can't be read as old or new style (hdf5) format")
            else:
                raise IOError('Input file '+filename+" can't be read - it could be a new style (hdf5) format, which requires the hdf5storage module to be installed.")
    if station_distribution:
        return _parse_matlab_station_distribution(data)
    return _parse_matlab_output(data)


def _parse_matlab_stations(stations_data):
    try:
        stations = {'names': stations_data[0]['Name'][0], 'azimuth': stations_data[0]['Azimuth'][0],
                    'takeoff_angle': stations_data[0]['TakeOffAngle'][0]}
        try:
            stations['polarity'] = stations_data[0]['Polarity'][0]
        except Exception:
            stations['polarity'] = np.zeros(len(stations['names']))
    except Exception:
        try:
            stations = {
                'names': [], 'azimuth': [], 'takeoff_angle': [], 'polarity': []}
            for st in stations_data:
                stations['names'].append(st[0][0])
                stations['azimuth'].append(st[1][0][0])
                stations['takeoff_angle'].append(st[2][0][0])
                try:
                    stations['polarity'].append(st[3][0][0])
                except Exception:
                    stations['polarity'].append(0)
        except Exception:
            try:
                stations = {
                    'names': [], 'azimuth': [], 'takeoff_angle': [], 'polarity': []}
                for st in stations_data:
                    stations['names'].append(st[0])
                    stations['azimuth'].append(st[1])
                    stations['takeoff_angle'].append(st[2])
                    try:
                        stations['polarity'].append(st[3])
                    except Exception:
                        stations['polarity'].append(0)
            except Exception:
                try:
                    stations = {
                        'names': [], 'azimuth': [], 'takeoff_angle': [], 'polarity': []}
                    for st in stations_data:
                        stations['names'].append(st[0][0])
                        stations['azimuth'].append(st[1][0][0])
                        stations['takeoff_angle'].append(st[2][0][0])
                        try:
                            stations['polarity'].append(st[3][0][0])
                        except Exception:
                            stations['polarity'].append(0)
                except Exception:
                    stations = {}
    try:
        stations['names'] = stations['names'].tolist()
    except Exception:
        pass
    return stations


def _parse_matlab_station_distribution(station_distribution_data):
    try:
        station_distribution = {'probability': station_distribution_data[
            'StationDistribution']['Probability'][0][0]}
        distribution = []
        for st in station_distribution_data['StationDistribution']['Distribution'][0][0][0]:
            st_sample = _parse_matlab_stations(st)
            try:
                st_sample.pop('polarity')
            except Exception:
                pass
            distribution.append(st_sample)
        station_distribution['distribution'] = distribution
    except Exception:
        station_distribution = convert_keys_from_unicode(
            station_distribution_data['StationDistribution'])
    return station_distribution


def _parse_matlab_output(data):
    """
    Converts the matlab structured data from the matlab file and returns an output data dictionary.

    Args
        data: dict matlab data.

    Returns
        (dict,dict): tuple of (events,data) dictionaries of output data.
    """
    n_events = 1
    if (isinstance(data['Events'], dict) and 'MTSpace' not in data['Events'].keys()) or (not isinstance(data['Events'], dict) and 'MTSpace' not in data['Events'].dtype.names):
        ev_ind = [u.replace('MTSpace', '') for u in data['Events'].dtype.names if 'MTSpace' in u]
        n_events = len(ev_ind)
    if n_events > 1:
        # Multiple Events
        events = []
        stations = []
        for ind in ev_ind:
            ev_data = {}
            key_map = dict([(u, u[:-1].replace('_', ''))
                            for u in data['Events'].dtype.names if u[-1] == ind])
            key_map['ln_pdf'] = 'ln_pdf'
            key_map['Probability'] = 'Probability'
            for i, key in enumerate(key_map.keys()):

                if key_map[key] not in ['MTSpace']:
                    ev_data[key_map[key]] = data['Events'][key][0, 0].flatten()
                    if len(ev_data[key_map[key]]) == 1:
                        ev_data[key_map[key]] = ev_data[key_map[key]][0]
                else:
                    ev_data[key_map[key]] = data['Events'][key][0, 0]
            try:
                st = _parse_matlab_stations(data['Stations'][0, int(ind)-1])
            except Exception:
                st = {}
            events.append(ev_data)
            stations.append(st)
        return events, stations
    # Single Event
    try:
        stations = _parse_matlab_stations(data['Stations'])
    except Exception:
        stations = {}
    events = {}
    try:
        for key in data['Events'].dtype.fields.keys():

            if key not in ['MTSpace']:
                events[key] = data['Events'][key][0, 0].flatten()
                if len(events[key]) == 1:
                    events[key] = events[key][0]
            else:
                events[key] = data['Events'][key][0, 0]
    except Exception:
        events = convert_keys_from_unicode(data['Events'])
    if events['Probability'].max() == 0:
        events['Probability'] = np.exp(events['ln_pdf']-events['ln_pdf'].max())
    return (events, stations)


def read_pickle_output(filename, station_distribution=False):
    """
    Reads pickle output and returns the data

    Args
        filename: str filename to read.

    Returns
        (dict,dict): tuple of (events,data) dictionaries of output data.
    """
    with open(filename, 'rb') as f:
        data = pickle.load(f)
    if station_distribution:
        station_distribution = data['StationDistribution']
        station_distribution['probability'] = station_distribution.pop('Probability')
        old_dist = station_distribution.pop('Distribution')
        station_distribution['distribution'] = []
        for st in old_dist:
            station_distribution['distribution'].append({'azimuth': st['Azimuth'],
                                                         'takeoff_angle': st['TakeOffAngle'],
                                                         'names': st['Name']})
        return station_distribution
    stations = {}
    if 'Stations' in data.keys():
        stations = _parse_matlab_stations(data.pop('Stations'))
    events = data
    if 'Events' in data.keys():
        events = data.pop('Events')
    return (events, stations)


def read_hyp_output(filename):
    """
    Reads hyp output and returns the data

    Args
        filename: str filename to read.

    Returns
        (dict,dict): tuple of (events,data) dictionaries of output data.
    """
    try:
        hyp_data = parse_hyp(os.path.splitext(filename)[0]+'.hyp')
    except Exception:
        hyp_data = {}
    events = read_binary_output(os.path.splitext(filename)[0]+'.mt')
    key_map = {'takeoffangle': 'takeoff_angle'}
    try:
        stations = hyp_data['PPolarity']['Stations']
        stations['polarity'] = hyp_data['PPolarity']['Measured']
        for key in stations:
            new_key = key.lower()
            if key.lower() in key_map.keys():
                new_key = key_map[key.lower()]
            stations[new_key] = stations.pop(key)
    except Exception:
        stations = {}
    return events, stations


def read_scatangle_output(filename, number_location_samples=0, bin_size=0, **kwargs):
    """
    Reads scatangle file for plotting

    Args
        filename: str filename to read

    Keyword Args
        number_location_samples: int number of location samples to sub-sample to.
        bin_size: float bin size for binning the angle samples in.

    Returns
        dict: station distribution
    """
    data = {'StationDistribution': {}}
    data['StationDistribution']['Distribution'], data['StationDistribution']['Probability'] = parse_scatangle(filename,
                                                                                                              number_location_samples=number_location_samples,
                                                                                                              bin_size=bin_size)
    station_distribution = data['StationDistribution']
    station_distribution['probability'] = station_distribution.pop('Probability')
    old_dist = station_distribution.pop('Distribution')
    station_distribution['distribution'] = []
    for st in old_dist:
        station_distribution['distribution'].append({'azimuth': st['Azimuth'], 'takeoff_angle': st['TakeOffAngle'], 'names': st['Name']})
    return station_distribution

# Dictionary conversion functions for file_io


def convert_keys_to_unicode(dictionary,):
    """
    Converts dictionary keys to unicode

    Required for MATLAB -v7.3 format output. This recursively changes strings in a dictionary to unicode.

    Args
        dictionary: Input dictionary.
    Returns
        dictionary: Converted dictionary.
    """
    if sys.version_info.major > 2:
        return dictionary
    if isinstance(dictionary, list):
        new_list = []
        for item in dictionary:
            new_list.append(convert_keys_to_unicode(item))
        return new_list
    elif isinstance(dictionary, dict):
        new_dictionary = {}
        for key, value in dictionary.items():
            if isinstance(value, list):
                new_value = []
                for item in value:
                    new_value.append(convert_keys_to_unicode(item))
                value = new_value
            if isinstance(value, dict):
                new_dictionary[key.decode('unicode-escape')] = convert_keys_to_unicode(value)
            else:
                new_dictionary[key.decode('unicode-escape')] = convert_keys_to_unicode(value)
        return new_dictionary
    return dictionary


def convert_keys_from_unicode(dictionary,):
    """
    Converts dictionary keys from  unicode

    Required for MATLAB -v7 format output. This recursively changes strings in a dictionary from unicode.

    Args
        dictionary: Input dictionary.
    Returns
        dictionary: Converted dictionary.
    """
    if sys.version_info.major > 2:
        return dictionary
    if isinstance(dictionary, list):
        new_list = []
        for item in dictionary:
            new_list.append(convert_keys_from_unicode(item))
        return new_list
    elif isinstance(dictionary, dict):
        new_dictionary = {}
        for key, value in dictionary.items():
            if isinstance(value, list):
                new_value = []
                for item in value:
                    new_value.append(convert_keys_from_unicode(item))
                value = new_value
            if isinstance(value, dict):
                new_dictionary[key.encode('utf-8')] = convert_keys_from_unicode(value)
            else:
                new_dictionary[key.encode('utf-8')] = convert_keys_from_unicode(value)
        return new_dictionary
    return dictionary


def unique_columns(data, counts=False, index=False):
    """Get unique columns in an array (used for max probability MT)"""
    output = []
    if float('.'.join(np.__version__.split('.')[:2])) >= 1.13:
        unique_results = np.unique(data, return_index=index, return_counts=counts, axis=1)
        if isinstance(unique_results, tuple):
            unique = unique_results[0]
        else:
            unique = unique_results
        if index:
            idx = unique_results[1]
        if counts:
            unique_counts = unique_results[-1]
    else:
        data = np.array(data).T
        ind = np.lexsort(data.T)
        data = np.squeeze(data)
        if len(ind.shape) > 1:
            ind = np.squeeze(ind)
        unique = data[ind[np.concatenate(([True], np.any(data[ind[1:]] != data[ind[:-1]], axis=1)))]].T
        if counts:
            indx = np.nonzero(np.concatenate(([True], np.any(data[ind[1:]] != data[ind[:-1]], axis=1))))[0]
            unique_counts = []
            for u, j in enumerate(indx):
                try:
                    unique_counts.append(indx[u+1]-j)
                except Exception:
                    unique_counts.append(1)
        if index:
            idx = ind[np.concatenate(([True], np.any(data[ind[1:]] != data[ind[:-1]], axis=1)))]
    output.append(np.matrix(unique))
    if counts:
        output.append(np.array(unique_counts))
    if index:
        output.append(idx)
    if len(output) == 1:
        return output[0]
    else:
        return tuple(output)
